"""

Read: https://www.pgcon.org/2014/schedule/attachments/330_postgres-for-the-wire.pdf
"""
import logging
from dataclasses import dataclass
from typing import Callable, Set, Tuple, Any

from pypsql_api.wire.actions import write_ssl_resp_no, read_startup_message, read_plain_text_password_request, \
    write_plain_text_password_request, read_ssl_request, write_ready_for_query, write_auth_ok, read_receive_command, \
    write_empty_response, write_data_frame_response, read_receive_extended_command, write_ssl_resp_yes
from pypsql_api.wire.actions_types import Names
from pypsql_api.wire.errors.core import log_exception
from pypsql_api.wire.query.extended_query import process_parse_command, process_bind_command, process_execute_command, \
    read_till_sync, write_extended_data_frame_response, write_extended_empty_response, describe_statement_or_portal
from pypsql_api.wire.query.simple_query import process_simple_query


@dataclass
class State:
    allowed: Set[str]
    name: str
    fn: Callable[[Any], Tuple[Any, str]]

    def move_next(self, context: Any):
        return self.fn(context)


def move_state(context: Any, state: State) -> Tuple[Any, str]:
    ctx_new, new_state = state.move_next(context)

    if not new_state:
        return ctx_new, Names.CLOSE

    if not new_state in state.allowed:
        raise Exception(f"Invalid state {new_state} from {state.name}, allowed states are {state.allowed}")

    return ctx_new, new_state


processes = {
    Names.READ_SSL_REQUEST: State(name=Names.READ_SSL_REQUEST,
                                  allowed={Names.WRITE_SSL_NO, Names.WRITE_SSL_YES,
                                           Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST, Names.CLOSE}, fn=read_ssl_request),
    Names.WRITE_SSL_NO: State(name=Names.READ_SSL_REQUEST, allowed={Names.READ_STARTUP_MESSAGE, Names.CLOSE},
                              fn=write_ssl_resp_no),
    Names.WRITE_SSL_YES: State(name=Names.READ_SSL_REQUEST, allowed={Names.READ_STARTUP_MESSAGE, Names.CLOSE},
                               fn=write_ssl_resp_yes),

    Names.READ_STARTUP_MESSAGE: State(name=Names.READ_STARTUP_MESSAGE,
                                      allowed={Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST, Names.CLOSE,
                                               Names.READ_STARTUP_MESSAGE}, fn=read_startup_message),

    Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST: State(name=Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST,
                                                   allowed={Names.READ_PLAIN_TEXT_PASSWORD, Names.CLOSE},
                                                   fn=write_plain_text_password_request),

    Names.READ_PLAIN_TEXT_PASSWORD: State(name=Names.READ_PLAIN_TEXT_PASSWORD,
                                          allowed={Names.CLOSE, Names.WRITE_AUTH_OK},
                                          fn=read_plain_text_password_request),

    Names.WRITE_AUTH_OK: State(name=Names.READ_PLAIN_TEXT_PASSWORD,
                               allowed={Names.READY_FOR_QUERY, Names.CLOSE},
                               fn=write_auth_ok),

    Names.READY_FOR_QUERY: State(name=Names.READY_FOR_QUERY,
                                 allowed={Names.CLOSE, Names.RECEIVE_COMMAND}, fn=write_ready_for_query),

    Names.RECEIVE_COMMAND: State(name=Names.RECEIVE_COMMAND,
                                 allowed={Names.CLOSE, Names.READY_FOR_QUERY, Names.SIMPLE_QUERY,
                                          Names.PARSE},
                                 fn=read_receive_command),

    Names.SIMPLE_QUERY: State(name=Names.SIMPLE_QUERY,
                              allowed={Names.CLOSE, Names.READY_FOR_QUERY, Names.WRITE_EMPTY_RESPONSE,
                                       Names.WRITE_DATA_FRAME},
                              fn=process_simple_query),

    Names.WRITE_EMPTY_RESPONSE: State(name=Names.WRITE_EMPTY_RESPONSE,
                                      allowed={Names.CLOSE, Names.READY_FOR_QUERY},
                                      fn=write_empty_response),

    Names.WRITE_DATA_FRAME: State(name=Names.WRITE_DATA_FRAME,
                                  allowed={Names.CLOSE, Names.READY_FOR_QUERY},
                                  fn=write_data_frame_response),

    Names.PARSE: State(name=Names.PARSE,
                       allowed={Names.CLOSE, Names.ERROR, Names.SYNC, Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                       fn=process_parse_command),

    Names.BIND: State(name=Names.BIND,
                      allowed={Names.CLOSE, Names.ERROR,
                               Names.EXECUTE, Names.SYNC,
                               Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                      fn=process_bind_command),

    Names.EXECUTE: State(name=Names.EXECUTE,
                         allowed={Names.CLOSE, Names.ERROR,
                                  Names.SYNC,
                                  Names.WRITE_EXTENDED_EMPTY_RESPONSE,
                                  Names.WRITE_EXTENDED_DATA_FRAME,
                                  Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                         fn=process_execute_command),

    Names.SYNC: State(name=Names.SYNC,
                      allowed={Names.CLOSE, Names.READY_FOR_QUERY, Names.SYNC,
                               Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                      fn=read_till_sync),

    Names.WRITE_EXTENDED_EMPTY_RESPONSE: State(name=Names.WRITE_EXTENDED_EMPTY_RESPONSE,
                                               allowed={Names.CLOSE, Names.SYNC,
                                                        Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                                               fn=write_extended_empty_response),

    Names.WRITE_EXTENDED_DATA_FRAME: State(name=Names.WRITE_EXTENDED_DATA_FRAME,
                                           allowed={Names.CLOSE, Names.SYNC,
                                                    Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                                           fn=write_extended_data_frame_response),

    Names.DESCRIBE: State(name=Names.DESCRIBE,
                          allowed={Names.CLOSE, Names.SYNC,
                                   Names.WRITE_EXTENDED_DATA_FRAME, Names.WRITE_EXTENDED_EMPTY_RESPONSE,
                                   Names.RECEIVE_EXTENDED_QUERY_COMMAND},
                          fn=describe_statement_or_portal),

    Names.RECEIVE_EXTENDED_QUERY_COMMAND: State(name=Names.RECEIVE_EXTENDED_QUERY_COMMAND,
                                                allowed={Names.CLOSE, Names.EXECUTE, Names.BIND, Names.SYNC,
                                                         Names.PARSE, Names.DESCRIBE},
                                                fn=read_receive_extended_command),

}


def run_states(context, state_name):
    while state_name != Names.CLOSE:
        context, state_name = run_state(context, state_name)

    logging.error("End run states")

    return context, state_name


def run_state(context, state_name):
    logging.error(f"Running state {state_name}")
    state = processes.get(state_name)

    if hasattr(context.output, 'flush'):
        try:
            context.output.flush()
        except Exception as e:
            log_exception(e)

    context, state_name = move_state(context, state)

    if not state_name:
        state_name = Names.CLOSE

    return context, state_name



import io
import logging
from dataclasses import dataclass
from typing import Any, Optional, Tuple, List

from pypsql_api.wire.actions_types import Names
from pypsql_api.wire.bytes import ReadingIO


def parse_portal_name(portal_name: str) -> str:
    if not portal_name:
        return "UNNAMED"

    return portal_name.lower().strip()


@dataclass
class SSLRequest:
    request_code: int
    inner_buff: Optional[ReadingIO]

    @staticmethod
    def read(buff: ReadingIO):
        inner_buff, message_len = buff.read_int32_delim_message()
        return SSLRequest(request_code=inner_buff.read_int32(), inner_buff=inner_buff)


@dataclass
class StartupMessage:
    protocol_version: int
    user: str
    database: str
    options: str

    @staticmethod
    def read(buff: ReadingIO):
        logging.error(f"Reading startup message using {buff}")

        inner_buff, message_len = buff.read_int32_delim_message()

        if not inner_buff:
            raise Exception("Invalid startup message, the length must be > 4 bytes")

        protocol_version = inner_buff.read_int32()
        message_len -= 4
        d = {}

        if message_len > 0:
            while True:
                n = inner_buff.read_cstring()
                if not n:
                    break
                v = inner_buff.read_cstring()
                if not v:
                    break

                d[n] = v

        return StartupMessage(
            protocol_version=protocol_version, user=d.get('user', ''), database=d.get('database', ''),
            options=d.get('options', '')
        )


@dataclass
class PasswordMessage:
    password: str

    @staticmethod
    def read_body(buff: ReadingIO):
        return PasswordMessage(password=buff.read_cstring())


@dataclass
class QueryMessage:
    query: str

    process_name: Any = Names.SIMPLE_QUERY

    @staticmethod
    def read_body(buff: ReadingIO):
        return QueryMessage(query=buff.read_cstring())


@dataclass
class BindMessage:
    portal_name: str
    prepared_statement_name: str

    # If empty the default text parameter format code is applied
    # The parameter format codes. Each must presently be zero (text) or one (binary).
    format_codes: List[int]

    parameter_values: List[Optional[bytes]]

    # The result-column format codes. Each must presently be zero (text) or one (binary).
    result_format_codes: List[int]

    process_name: Any = Names.BIND

    @staticmethod
    def read_body(buff: ReadingIO):
        portal_name = parse_portal_name(buff.read_cstring())

        prepared_statement_name = parse_portal_name(buff.read_cstring())

        # read the parameter format codes
        number_of_format_codes = buff.read_int16()
        format_codes = [buff.read_int16() for _ in range(number_of_format_codes)]

        # read the parameter values
        number_of_params = buff.read_int16()
        parameter_values = []
        for _ in range(number_of_params):
            param_len = buff.read_int32()
            if param_len == 0 or param_len == -1:
                parameter_values.append(None)
            else:
                parameter_values.append(buff.read_bytes(param_len))

        number_of_result_format_codes = buff.read_int16()
        result_format_codes = [buff.read_int16() for _ in range(number_of_result_format_codes)]

        return BindMessage(
            portal_name=portal_name,
            prepared_statement_name=prepared_statement_name,
            format_codes=format_codes,
            parameter_values=parameter_values,
            result_format_codes=result_format_codes
        )


@dataclass
class ParseMessage:
    name: str
    query: str

    # zero is unspecified
    oid_types: List[int]

    process_name: Any = Names.PARSE

    @staticmethod
    def read_body(buff: ReadingIO):
        assert isinstance(buff, ReadingIO), f"got {buff}"
        name = parse_portal_name(buff.read_cstring())

        query = buff.read_cstring()

        parameter_data_types_len = buff.read_int16()

        oid_types = [buff.read_int32() for _ in range(parameter_data_types_len)]

        logging.info(f"Parse : name {name}, {query}, {oid_types}")

        return ParseMessage(
            name=name,
            query=query,
            oid_types=oid_types
        )


@dataclass
class ExecuteMessage:
    portal_name: str
    max_rows: int

    process_name: Any = Names.EXECUTE

    def unlimited(self):
        return self.max_rows == 0

    @staticmethod
    def read_body(buff: ReadingIO):
        return ExecuteMessage(portal_name=parse_portal_name(buff.read_cstring()),
                              max_rows=buff.read_int32())


@dataclass
class Terminate:
    process_name: Any = Names.CLOSE

    @staticmethod
    def read_body(_: ReadingIO):
        return Terminate()


@dataclass
class Flush:
    process_name: Any = Names.FLUSH

    @staticmethod
    def read_body(_: ReadingIO):
        return Flush()


@dataclass
class Sync:
    process_name: Any = Names.SYNC

    @staticmethod
    def read_body(_: ReadingIO):
        return Sync()


@dataclass
class Describe:
    t: str  # 'S == describe prepared statement, P == describe portal
    name: str

    process_name: Any = Names.DESCRIBE

    @staticmethod
    def read_body(buff: ReadingIO):
        t = buff.read_byte()
        name = buff.read_cstring()
        if not name:
            name = 'UNNAMED'

        return Describe(t=t, name=name)


message_type_map = {
    ord('p'): PasswordMessage.read_body,
    ord('Q'): QueryMessage.read_body,
    ord('X'): Terminate.read_body,
    ord('P'): ParseMessage.read_body,
    ord('B'): BindMessage.read_body,
    ord('E'): ExecuteMessage.read_body,
    ord('H'): Flush.read_body,
    ord('S'): Sync.read_body,
    ord('D'): Describe.read_body,
}


class Message:

    @staticmethod
    def read(buff: ReadingIO) -> Tuple[Optional[Any], Optional[bytes]]:
        """
        Returns [Optional[Message], type:bytes
        If The message and Type are None, we've reached the end of stream
        """

        t = buff.read_byte()

        if not t:
            return None, None

        fn = message_type_map.get(ord(t), None)
        logging.info(f"Read message type {t} using fn {fn}")
        if fn:
            try:
                inner_buff, _ = buff.read_int32_delim_message()
                return fn(inner_buff), t
            except Exception as e:
                logging.error(e)
                return fn(ReadingIO(io.BytesIO())), t

        return None, t



import logging
import traceback

import ssl
from io import BytesIO
from typing import Optional

from pypsql_api.config.types import Session
from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context
from pypsql_api.wire.back import SSLNo, AuthenticationCleartextPassword, ReadyForQuery, AuthenticationOk, \
    EmptyQueryResponse, DataFrameRowDescription, DataFrameDataRows, CommandComplete, SSLYes, ErrorResponse, \
    ParameterStatus
from pypsql_api.wire.bytes import ReadingIO, WritingIO
from pypsql_api.wire.errors.core import log_exception, NoData
from pypsql_api.wire.front import SSLRequest, StartupMessage, Message, PasswordMessage

cipher_list = "ECDH+ECDSA+AESGCM:" \
 \
              "ECDH+ECDSA+AES:" \
              "ECDH+AESGCM:" \
              "ECDH+AES:" \
 \
              "DH+ECDSA+AESGCM:" \
              "DH+ECDSA+AES:" \
              "DH+AESGCM:" \
              "DH+AES:" \
 \
              "RSA+AESGCM:" \
              "RSA+AESCBC:" \
 \
              "!aNULL:!MD5:!DSS:"


def reset(rio: ReadingIO) -> ReadingIO:
    rio.buff.seek(0)
    return rio


def prefix_len(rio: ReadingIO) -> ReadingIO:
    bts = rio.read_bytes()
    wio = WritingIO(BytesIO())
    wio.write_int32(len(bts))
    wio.write_bytes(bts)

    buff = wio.buff
    buff.seek(0)

    return ReadingIO(buff)


def read_ssl_request(context: Context):
    logging.error("read_ssl_requst")
    try:
        # Add support for ancient ssl protocol
        # New protocol SSLReq, SSLAns, StartupMsg
        # Ancient: StartupMsg

        ssl_request = SSLRequest.read(context.input)
        logging.error(f"read ssl request {ssl_request}")

        if ssl_request.request_code == 196608:
            # ancient mode
            logging.error("reading startup message from ancient")
            return read_startup_message(context, inner_buff=prefix_len(reset(ssl_request.inner_buff)))
        else:

            if context.key_file and context.cert_file:
                return context.update_mem('ssl_request', ssl_request), Names.WRITE_SSL_YES
            else:
                return context.update_mem('ssl_request', ssl_request), Names.WRITE_SSL_NO

    except Exception as e:
        log_exception(e)
        logging.error("Error while processing ssl request, downgrading to SSL_NO")
        return context, Names.CLOSE


def write_ssl_resp_no(context: Context):
    SSLNo().write(context.output)

    return context, Names.READ_STARTUP_MESSAGE


def write_ssl_resp_yes(context: Context):
    SSLYes().write(context.output)

    # we need wrap the connection in a SSL connection
    ssl_obj: ssl.SSLSocket = ssl.wrap_socket(sock=context.socket,
                                             do_handshake_on_connect=False,
                                             server_side=True,
                                             certfile=context.cert_file,
                                             keyfile=context.key_file,
                                             ciphers=cipher_list)

    ssl_obj.do_handshake()

    context.mem['ssl_obj'] = ssl_obj

    context.input = ReadingIO(ssl_obj)
    context.output = WritingIO(ssl_obj)

    return context, Names.READ_STARTUP_MESSAGE


def read_startup_message(context: Context, inner_buff: Optional[ReadingIO] = None):
    try:

        if inner_buff is None:
            io = context.input
        else:
            logging.error("Using inner buffer to read the startup message")
            io = inner_buff

        startup_front = StartupMessage.read(io)

        logging.error(f"Got startup message: {startup_front}")

        session = Session(
            user=startup_front.user, database=startup_front.database, password=''
        )

        context.session = session
        return context.update_mem('session', session), Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST
    except NoData as e:
        log_exception(e)
        return context, Names.CLOSE
    except Exception as e:
        log_exception(e)
        raise


def write_plain_text_password_request(context: Context):
    AuthenticationCleartextPassword().write(context.output)
    context.output.flush()

    return context, Names.READ_PLAIN_TEXT_PASSWORD


def read_plain_text_password_request(context: Context):
    m, t = Message.read(context.input)

    if not (m and t):
        return context, None

    if not isinstance(m, PasswordMessage):
        raise Exception(f"UnExpected message {m}")

    session = context.session
    session.password = m.password
    auth_ok, msg = context.auth_handler.handle(session=session)

    if auth_ok:
        return context, Names.WRITE_AUTH_OK
    else:
        ErrorResponse.severe("Password not correct").write(context.output)
        return context, Names.CLOSE


def write_ready_for_query(context: Context):
    ReadyForQuery().write(context.output)
    context.output.flush()

    return context, Names.RECEIVE_COMMAND


def write_auth_ok(context: Context):
    AuthenticationOk().write(context.output)

    # write out the parameters for the server
    for k,v in context.parameters.items():
        ParameterStatus(name=k, value=v).write(context.output)

    context.output.flush()

    return context, Names.READY_FOR_QUERY


def read_receive_command(context: Context):
    m, t = Message.read(context.input)

    logging.info(f">>read_receive_command msg {m}, {t}")
    if not (m and t):
        return context, None

    return context.update_mem('message', m), m.process_name


def read_receive_extended_command(context: Context):
    m, t = Message.read(context.input)

    logging.info(f">>read_receive_extended_command msg {m}, {t}")
    if m is None and t is None:
        logging.info(f"Not m,t  {m}, {t}")
        return context, None

    if m and m.process_name in {Names.EXECUTE, Names.BIND, Names.SYNC, Names.DESCRIBE}:
        return context.update_mem('message', m), m.process_name

    logging.info(f"Expected an extended protocol query message but got {m}, {t}")
    return context, Names.ERROR


def write_empty_response(context: Context):
    EmptyQueryResponse().write(context.output)

    return context, Names.READY_FOR_QUERY


def write_data_frame_response(context: Context):
    df = context.mem['data']
    if df is None:
        raise Exception("We expect a data frame instance here")

    DataFrameRowDescription(df=df).write(context.output)
    rows = DataFrameDataRows(df=df, offset=0, max_rows=1000000).write(context.output)

    CommandComplete(tag=f"SELECT {rows}").write(context.output)

    context.output.flush()

    return context, Names.READY_FOR_QUERY


def read_parse_command(context: Context):
    m, t = Message.read(context.input)



import enum


class Names(enum.Enum):
    ERROR = 0
    READ_SSL_REQUEST = 1
    WRITE_SSL_YES = 2
    WRITE_SSL_NO = 3

    WRITE_PLAIN_TEXT_PASSWORD_REQUEST = 4
    READ_PLAIN_TEXT_PASSWORD = 5

    READ_STARTUP_MESSAGE = 6

    CLOSE = 7

    READY_FOR_QUERY = 8
    WRITE_AUTH_OK = 9
    RECEIVE_COMMAND = 10  # after ready for query we wait for the next command
    SIMPLE_QUERY = 11

    WRITE_EMPTY_RESPONSE = 12
    WRITE_DATA_FRAME = 13

    PARSE = 14
    BIND = 15
    EXECUTE = 16
    SYNC = 17
    FLUSH = 18
    RECEIVE_EXTENDED_QUERY_COMMAND = 19  # after ready for query we wait for the next command
    WRITE_EXTENDED_EMPTY_RESPONSE = 20
    WRITE_EXTENDED_DATA_FRAME = 21

    DESCRIBE = 22


import logging
import traceback


class NoData(Exception):
    """
    No data on the TCP connection, or less than expected
    """

    def __init__(self, expected_bytes: int, got_bytes: int, *args, **kwargs):
        super(Exception, self).__init__(*args, **kwargs)
        self.expected_bytes = expected_bytes
        self.got_bytes = got_bytes


def log_exception(e: Exception):
    try:
        logging.exception(e)
    except:
        pass



import io
import logging
from dataclasses import dataclass
from typing import Any, List

from numpy import dtype
from pandas import DataFrame

from pypsql_api.wire.bytes import WritingIO
from pypsql_api.wire.dataframe.types import TYPE_MAP, DataType


@dataclass
class ParameterStatus:
    name: str
    value: str

    def write(self, buff: WritingIO):
        buff.write_byte(ord('S'))
        internal_buff = WritingIO(io.BytesIO())

        internal_buff.write_cstring(self.name)
        internal_buff.write_cstring(self.value)

        bytes = internal_buff.buff.getvalue()

        buff.write_int32(len(bytes) + 4)
        buff.write_bytes(bytes)


@dataclass
class ErrorField:
    type: int  # one of https://www.postgresql.org/docs/current/protocol-error-fields.html
    value: Any


@dataclass
class ErrorResponse:
    fields: List[ErrorField]

    def write(self, buff: WritingIO):
        buff.write_byte(ord('E'))
        internal_buff = WritingIO(io.BytesIO())
        for field in self.fields:
            internal_buff.write_byte(field.type)
            internal_buff.write_cstring(field.value)

        bytes = internal_buff.buff.getvalue()

        buff.write_int32(len(bytes) + 5)
        buff.write_bytes(bytes)

        buff.write_byte(0)

    @staticmethod
    def severe(msg: str):
        """
        Creates a simple string error response
        """
        return ErrorResponse(fields=[ErrorField(type=ord('M'), value=msg)])

    @staticmethod
    def query(msg: str):
        """
        Creates a simple string error response
        """
        return ErrorResponse(fields=[ErrorField(type=ord('W'), value=msg)])


@dataclass
class ReadyForQuery:
    transaction_indicator: int = ord('I')

    def write(self, buff: WritingIO):
        buff.write_byte(ord('Z'))
        buff.write_int32(5)
        buff.write_byte(self.transaction_indicator)


class ParseComplete:

    def write(self, buff: WritingIO):
        buff.write_byte(ord('1'))
        buff.write_int32(4)


class BindComplete:

    def write(self, buff: WritingIO):
        buff.write_byte(ord('2'))
        buff.write_int32(4)


class PortalSuspended:

    def write(self, buff: WritingIO):
        buff.write_byte(ord('s'))
        buff.write_int32(4)


class AuthenticationOk:

    def write(self, buff: WritingIO):
        buff.write_byte(ord('R'))
        buff.write_int32(8)
        buff.write_int32(0)


class AuthenticationCleartextPassword:

    def write(self, buff: WritingIO):
        buff.write_byte(ord('R'))
        buff.write_int32(8)
        buff.write_int32(3)


class SSLYes:
    def write(self, buff: WritingIO):
        buff.write_byte(ord('S'))


class SSLNo:
    def write(self, buff: WritingIO):
        buff.write_byte(ord('N'))


class EmptyQueryResponse:

    def write(self, buff: WritingIO):
        buff.write_byte(ord('I'))
        buff.write_int32(4)


@dataclass
class CommandComplete:
    tag: str

    def write(self, buff: WritingIO):
        buff.write_byte(ord('C'))
        internal_buff = WritingIO(io.BytesIO())
        internal_buff.write_cstring(self.tag)

        bytes = internal_buff.buff.getvalue()

        buff.write_int32(len(bytes) + 4)  # Length of message contents in bytes, including self.
        buff.write_bytes(bytes)
        logging.info("writing COMMAND COMPLETE")
        logging.info(self.tag)


def is_nan(x):
    return x != x


@dataclass
class DataFrameDataRows:
    df: DataFrame
    offset: int
    max_rows: int

    def write_cell(self, buff: WritingIO, row_i: int, col_name, type_map: DataType, df: DataFrame):
        """
        Write out a single column cell value
        """
        v = df.iloc[row_i][col_name]

        if v is None or is_nan(v):
            buff.write_int32(-1)
        else:
            bts = type_map.typsend(v)

            buff.write_int32(len(bts))  # type length in bytes
            buff.write_bytes(bts)

    def write(self, buff: WritingIO):

        df = self.df

        columns = dict(df.dtypes)
        column_names = list(columns.keys())
        row_count = len(df.index)

        column_types = [TYPE_MAP[dt.type] for _, dt in columns.items()]

        offset = self.offset
        if offset < row_count:
            for row_i in range(row_count):
                # for each row
                # write a DataRow
                buff.write_byte(ord('D'))  # Identifies the message as a data row.

                internal_buff = WritingIO(io.BytesIO())
                internal_buff.write_int16(len(column_types))

                for i, type_map in enumerate(column_types):
                    self.write_cell(internal_buff, row_i + offset, column_names[i], type_map, df)

                data_row_bts = internal_buff.buff.getvalue()

                buff.write_int32(len(data_row_bts) + 4)  # Length of message contents in bytes, including self.
                buff.write_bytes(data_row_bts)

            logging.info(f"done writing datarows {row_count}")
            return row_count

        return 0


@dataclass
class DataFrameRowDescription:
    df: DataFrame

    def write_field_desc(self, buff: WritingIO, n: Any, dt: dtype):
        # see https://github.com/postgres/postgres/blob/master/src/include/catalog/pg_type.dat
        logging.info(f"write field {n} = [{dt}, {dt.type}]")

        type_map: DataType = TYPE_MAP[dt.type]

        buff.write_int32(
            0)  # If the field can be identified as a column of a specific table, the object ID of the table; otherwise zero.
        buff.write_int16(
            0)  # If the field can be identified as a column of a specific table, the attribute number of the column; otherwise zero.
        buff.write_int32(int(type_map.oid))  # The object ID of the field's data type.

        buff.write_int16(int(type_map.typlen))  # The data type size (see pg_type.typlen)
        buff.write_int32(-1)  # The type modifier
        # zero is text format for all
        buff.write_int16(0)  # keep always as text, only binary for FETCH commands

    def write(self, buff: WritingIO):
        buff.write_byte(ord('T'))  # Identifies the message as a row description.

        internal_buff = WritingIO(io.BytesIO())

        df = self.df

        columns = dict(df.dtypes)

        internal_buff.write_int16(len(columns))  # Specifies the number of fields in a row (can be zero)

        for n, dt in columns.items():
            internal_buff.write_cstring(n)  # The field name.
            self.write_field_desc(internal_buff, n, dt)

        bytes = internal_buff.buff.getvalue()

        buff.write_int32(len(bytes) + 4)  # Length of message contents in bytes, including self.
        buff.write_bytes(bytes)



from pandas import DataFrame

from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context
from pypsql_api.wire.back import CommandComplete


def process_simple_query(context: Context):
    msg = context.mem.get('message', None)

    if msg.query:
        msg.query = msg.query.replace('OPERATOR(pg_catalog.~)', '=') \
            .replace('~', '=')

        query_handler = context.query_handler

        query = msg.query.lower()

        if query.startswith("set"):
            query_handler.handle_set(session=context.session, sql=msg.query)
            CommandComplete(tag="SET").write(context.output)
            return context, Names.READY_FOR_QUERY
        elif query == 'begin':
            query_handler.handle_begin(session=context.session, sql=msg.query)
            CommandComplete(tag="BEGIN").write(context.output)
            return context, Names.READY_FOR_QUERY
        elif query == 'end':
            query_handler.handle_end(session=context.session, sql=msg.query)
            CommandComplete(tag="END").write(context.output)
            return context, Names.READY_FOR_QUERY
        elif query == 'rollback':
            query_handler.handle_rollback(session=context.session, sql=msg.query)
            CommandComplete(tag="ROLLBACK").write(context.output)
            return context, Names.READY_FOR_QUERY
        else:
            resp, _ = query_handler.handle(session=context.session, sql=msg.query)

            if isinstance(resp, DataFrame):
                return context.update_mem('data', resp), Names.WRITE_DATA_FRAME

            return context, Names.WRITE_EMPTY_RESPONSE



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypsql_api',
 'pypsql_api.config',
 'pypsql_api.ext',
 'pypsql_api.wire',
 'pypsql_api.wire.dataframe',
 'pypsql_api.wire.errors',
 'pypsql_api.wire.extended',
 'pypsql_api.wire.query']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.2.2,<2.0.0']

setup_kwargs = {
    'name': 'pypsql-api',
    'version': '0.2.4',
    'description': 'Add postgresql support to your data APIs',
    'long_description': None,
    'author': 'Gerrit Jansen van Vuuren',
    'author_email': 'gerritjvv@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Tuple, Optional, Union, Any

from pandas import DataFrame

from pypsql_api.config.types import Session
from pypsql_api.wire.extended.types import PreparedStatement, Portal
from pypsql_api.wire.front import ParseMessage, BindMessage


class AuthHandler:

    def handle(self, session: Session) -> Tuple[bool, Optional[str]]:
        """
        :param session:
        :return: Tuple[bool, Optional[str]] if (true, any) then AuthOK, otherwise AuthFail, and the message in the str
        contains any error message
        """
        raise Exception("Not implemented")


class QueryHandler:

    def handle_rollback(self, session: Session, sql):
        pass

    def handle_begin(self, session: Session, sql):
        pass

    def handle_end(self, session: Session, sql):
        pass

    def handle_set(self, session: Session, sql):
        pass

    def handle(self, session: Session, sql) -> Tuple[Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:
        """
        :param session:
        :param sql: any sql statement passed by the client,
         this can be multiple statements and being-commit statements.
        :return: None -> No result
                 (Any, Msg)  -> Msg for future use to return notification messages.
                 (DataFrame, Any) -> The DataFrame the contains the data and schema returned
        """
        raise Exception("Not implemented")


class ExtendedQueryHandler:
    """
    Parse prepared statement queries
    """

    def parse(self, session: Session, msg: ParseMessage) -> Tuple[
        Optional[Union[Any, Optional[PreparedStatement]]], Optional[Any]]:
        """
        :param session:
        :param msg: ParseMessage
         this can be multiple statements and being-commit statements.
        :return: None -> No result
                 (Any, Msg)  -> Msg for future use to return notification messages.
                 (PreparedStatement, Any) -> The DataFrame the contains the data and schema returned
        """
        raise Exception("Not implemented")

    def bind(self, session: Session, prepared_statement: PreparedStatement, bind: BindMessage) -> Tuple[
        Optional[Union[Any, Optional[Portal]]], Optional[Any]]:
        """
        :param session:
        :param prepared_statement: PreparedStatement
        :param bind: BindMessage
         this can be multiple statements and being-commit statements.
        :return: None -> No result
                 (Any, Msg)  -> Msg for future use to return notification messages.
                 (Portal, Any) -> The DataFrame the contains the data and schema returned
        """
        raise Exception("Not implemented")

    def execute(self, session: Session, portal: Portal, max_rows: int) -> Tuple[Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:
        """
        Any exception throws an error
        :param session:
        :param portal: Portal
        :param max_rows: int
        :return: None -> No result
                 (Any, Msg)  -> Msg for future use to return notification messages.
                 (PreparedStatement, Any) -> The DataFrame the contains the data and schema returned
        """
        raise Exception("Not implemented")
